/*******************************************
**** VIDEO.C - Video Hardware Emulation ****
*******************************************/

/*
* DEFINES:
* MODEONSCALE - define if you require video mode to be set
*               on window resize.
*               otherwise it will only be reset on switching to fullscreen.
* MODEREINIT  - define if you need to reinitialize OpenGL on setting video mode.
* SCALEREINIT - define if you need to reinitialize OpenGL on setting video mode.
*/

/*-- Include Files -----------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>

#include "SDL.h"
#include "SDL_opengl.h"

#include <string.h>
#include <math.h>

#include "video.h"
#include "video3d.h"
#include "pacifi3d.h"


/*-- Defines -----------------------------------------------------------------*/
/* width = 224, height= 288*/
#define SCREENWIDTH (28*8)
#define SCREENHEIGHT (36*8)
#define SCREENRATIO (((float)SCREENWIDTH)/((float)SCREENHEIGHT))

#define TEXTUREWIDTH 256
#define TEXTUREHEIGHT 512

#define MAXTEXX (float)( ((float)SCREENWIDTH) / ((float)TEXTUREWIDTH))
#define MAXTEXY (float)( ((float)SCREENHEIGHT) / ((float)TEXTUREHEIGHT))

#define TOTAL_COLORS 0x80

/*-- Imported Variables ------------------------------------------------------*/


/*-- Global Variables --------------------------------------------------------*/
Uint8 tilerom[0x1000];
Uint8 spriterom[0x1000];
Uint8 prom[0x120];
int colortable[0x80];

/* OpenGL palette*/
float glcolors[0x80][3];
float colors[0x10][3];

Uint32 paldata[0x10];
Uint32 paldata2[0x80];

Uint8 tiles[256][8][8];
Uint8 sprites[64][16][16];


SDL_Surface      *screen;
SDL_Surface      *video_buffer;
SDL_Surface      *tile_layer;

SDL_Rect         screenrect={0,0,SCREENWIDTH,SCREENHEIGHT};

int fullscreen = 0;
int flipscreen = 0;
int display_mode=THIRD_PERSON;
int snap_no=0;
GLuint pactexture;
GLuint displayList;

/* SDL interprets each pixel as a 32-bit number, so our masks must depend
   on the endianness (byte order) of the machine */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    Uint32 rmask = 0xff000000;
    Uint32 gmask = 0x00ff0000;
    Uint32 bmask = 0x0000ff00;
    Uint32 amask = 0x00000000;
#else
    Uint32 rmask = 0x000000ff;
    Uint32 gmask = 0x0000ff00;
    Uint32 bmask = 0x00ff0000;
    Uint32 amask = 0x00000000;
#endif

/* Flags we will pass into SDL_SetVideoMode. */
static int flags = SDL_OPENGL | SDL_RESIZABLE;
/* Color depth in bits of our window. */
static int bpp=32;
/* Dimensions of our window. */
static int width = 0;
static int height = 0;
static float ratio=1.0;

int magictile=255;


/*-- Function Prototypes -----------------------------------------------------*/
int     video_init(void);
void    snapshot_init(void);
void    video_palette_init(void);
void    video_tile_init(void);
void    video_sprite_init(void);

void    video_shutdown(void);
void    video_fullscreen_toggle(void);

void    incframeskip(void);

void    drawtile(int tile, int color, int flipx, int flipy, int x, int y);
void    video_draw_screen(void);
void    video_save_snapshot(void);
void    video_setup(void);

void setup_opengl( int w, int h );

void setup_texture(void);

/*----------------------------------------------------------------------------*/
void init_gl_objs(void) {
    setup_texture();
    init_pacman_3d_objs();
}


int    video_init(void)
{
    /* Information about the current video settings. */
    const SDL_VideoInfo* info = NULL;

    /* First, initialize SDL's video subsystem. */
    if( SDL_Init( SDL_INIT_VIDEO ) < 0 ) {
        /* Failed, exit. */
        fprintf( stderr, "Video initialization failed: %s\n", SDL_GetError( ) );
        exit( 1 );
    }

    /* Let's get some video information. */
    info = SDL_GetVideoInfo( );

    if( !info ) {
        /* This should probably never happen. */
        fprintf( stderr, "Video query failed: %s\n",  SDL_GetError( ) );
        exit( 1 );
    }

    width = 800;
    height = 600;
    bpp = info->vfmt->BitsPerPixel;

    /* setup opengl */
    SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
    SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
    SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
    SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

    /*
     * Set the video mode
     */
#ifndef MODEONSCALE
    if( (screen=SDL_SetVideoMode( width, height, bpp, flags )) == 0) {
        /*
         * This could happen for a variety of reasons,
         * including DISPLAY not being set, the specified
         * resolution not being available, etc.
         */
        fprintf( stderr, "Video mode set failed: %s\n",
             SDL_GetError( ) );
        exit( 1 );
    }
#endif
    setup_opengl(width, height);

    /* create 2D pixel buffers for pacman emulation */
    video_buffer = SDL_CreateRGBSurface(SDL_SWSURFACE, SCREENWIDTH, SCREENHEIGHT, 32,
                                        rmask, gmask, bmask, amask);
    tile_layer = SDL_CreateRGBSurface(SDL_SWSURFACE, SCREENWIDTH, SCREENHEIGHT, 32,
                                        rmask, gmask, bmask, amask);

    if(video_buffer == NULL) {
        fprintf(stderr, "CreateRGBSurface failed: %s\n", SDL_GetError());
        exit(1);
    }
    
    if(tile_layer == NULL) {
        fprintf(stderr, "CreateRGBSurface failed: %s\n", SDL_GetError());
        exit(1);
    }

    snapshot_init();

    video_palette_init();
    video_tile_init();
    video_sprite_init();

    /* initialise 3d objects */
    init_gl_objs();

    return 1;
}


void setup_opengl( int w, int h )
{
    ratio = (float) w / (float) h;

    /* store new width and height */
    width=w;
    height=h;

    /* resetting videomode on win32 causes opengl state to be lost */
    /* but its resizes scales ok without setting video mode */
#ifdef MODEONSCALE
    if( (screen=SDL_SetVideoMode( width, height, bpp, flags )) == 0) {
        /*
         * This could happen for a variety of reasons,
         * including DISPLAY not being set, the specified
         * resolution not being available, etc.
         */
        fprintf( stderr, "Video mode set failed: %s\n",
             SDL_GetError( ) );
        exit( 1 );
    }
#endif

    /* Our shading model--Gouraud (smooth). */
    glShadeModel( GL_SMOOTH );
    glEnable (GL_DEPTH_TEST);  /* enable depth buffering */
    glFrontFace( GL_CW );

    /* Set the clear color. */
    glClearColor( 0, 0, 0, 0 );
    glClearDepth(100.0);

    /* Setup our viewport. */
    glViewport( 0, 0, width, height );

    /*
     * Change to the projection matrix and set
     * our viewing volume.
     */
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );
    gluPerspective( 60.0, ratio, 1.0, 8192.0 );
    glTranslatef(0.0,0.0,-2.5);
    glMatrixMode( GL_MODELVIEW );

#ifdef SCALEREINIT
    /* recalculate everything */
    init_gl_objs();
#endif

}

/*----------------------------------------------------------------------------*/
void    video_shutdown(void)
{

    if (video_buffer != NULL) SDL_FreeSurface( video_buffer );
    if (tile_layer != NULL) SDL_FreeSurface( tile_layer );

}

/*----------------------------------------------------------------------------*/


void video_fullscreen_toggle(void)
{
    flags^=SDL_FULLSCREEN;
    /* but you do need to set the video mode to get fullscreen */

#ifndef MODEONSCALE
    if( (screen=SDL_SetVideoMode( 800, 600, bpp, flags )) == 0) {
        /*
         * This could happen for a variety of reasons,
         * including DISPLAY not being set, the specified
         * resolution not being available, etc.
         */
        fprintf( stderr, "Video mode set failed: %s\n",
             SDL_GetError( ) );
        exit( 1 );
    }
#endif
    setup_opengl(800,600);
    /* resetting videomode on win32 causes opengl state to be lost */
#ifdef MODEREINIT
    /* so we have to recalculate everything */
    init_gl_objs();
#endif

}


void video_mode_toggle(void)
{
    display_mode=(display_mode+1)%MAX_VIEWS;
    /* reset control direction if returning to non-first person style mode */
    if (display_mode < FIRST_PERSON) {
       input_direction(NORTH);
    }
}


/*----------------------------------------------------------------------------*/
void palette_set_color(int i, int red, int green, int blue) {
    colors[i][0]=((float)red)/255.0;
    colors[i][1]=((float)green)/255.0;
    colors[i][2]=((float)blue)/255.0;
    
    paldata[i]=SDL_MapRGB(video_buffer->format,red,green,blue);
}


void    video_palette_init(void)
{
    /* palette calc */
    int i;
    char * color_prom=prom;

    for (i = 0;i < 0x10;i++)
    {
        int bit0,bit1,bit2,r,g,b;


        /* red component */
        bit0 = (*color_prom >> 0) & 0x01;
        bit1 = (*color_prom >> 1) & 0x01;
        bit2 = (*color_prom >> 2) & 0x01;
        r = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;
        /* green component */
        bit0 = (*color_prom >> 3) & 0x01;
        bit1 = (*color_prom >> 4) & 0x01;
        bit2 = (*color_prom >> 5) & 0x01;
        g = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;
        /* blue component */
        bit0 = 0;
        bit1 = (*color_prom >> 6) & 0x01;
        bit2 = (*color_prom >> 7) & 0x01;
        b = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;

        palette_set_color(i,r,g,b);
        color_prom++;
    }

    color_prom += 0x10;
    /* color_prom now points to the beginning of the lookup table */

    /* character lookup table */
    /* sprites use the same color lookup table as characters */
    for (i = 0;i < TOTAL_COLORS;i++) {
        colortable[i] = *(color_prom++) & 0x0f;
	
	paldata2[i]=paldata[colortable[i]];
	
        glcolors[i][0] = colors[colortable[i]][0];
        glcolors[i][1] = colors[colortable[i]][1];
        glcolors[i][2] = colors[colortable[i]][2];
    }
}




/*---------------------------------------------------------------------------*/
/* tiles */

#if 0
static struct GfxLayout tilelayout =
{
    8,8,    /* 8*8 characters */
    256,    /* 256 characters */
    2,  /* 2 bits per pixel */
    { 0, 4 },   /* the two bitplanes for 4 pixels are packed into one byte */
    { 8*8+0, 8*8+1, 8*8+2, 8*8+3, 0, 1, 2, 3 }, /* bits are packed in groups of four */
    { 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8 },
    16*8    /* every char takes 16 bytes */
};
#endif


void video_tile_init (void) {
    int tile, x, y;

    for (tile=0;tile<256;tile++) {
        for (y=0; y<8; y++) {
            for (x=0; x<4; x++) {
                tiles[tile][x][y]=((tilerom[(tile*16)+y]>>x)&0x01) |
                                      ((tilerom[(tile*16)+y]>>(x+3))&0x02);

                tiles[tile][x+4][y]=((tilerom[(tile*16)+y+8]>>x)&0x01) |
                                      ((tilerom[(tile*16)+y+8]>>(x+3))&0x02);
            }
        }
    }
}



void drawtile(int tile, int color, int flipy, int flipx, int y, int x) {
    Uint32 *pixels;
    int i,j,ti;

    x=SCREENWIDTH-x-8;
    color*=4;

    if(SDL_MUSTLOCK(tile_layer))
    SDL_LockSurface(tile_layer);

    pixels=(Uint32*)tile_layer->pixels;
    pixels+=(y*SCREENWIDTH) + x;

    for (i=0;i<8;i++) {
	ti=flipy?i:(7-i);
        for (j=0;j<8;j++) {
           pixels[j]=paldata2[color + tiles[tile][ti][flipx?j:(7-j)]];
        }
        pixels+=SCREENWIDTH;
    }

    if(SDL_MUSTLOCK(tile_layer))
    SDL_UnlockSurface(tile_layer);

}



void video_inc_tile(void) {
  magictile=(magictile+1)%256;
  printf("magictile=%d\n",magictile);
}

void video_dec_tile(void) {
  magictile=(magictile-1);
  if (magictile==-1) magictile=255;
  printf("magictile=%d\n",magictile);
}


/* tiles */
/*  
    0 = zero
    16 = pill
    20 = power pill

    38 = Quote
    48 - 57 = zero - nine
    64 = transparent
    65 - 90 = A - Z

    93-95 = PTS (title screen)

    160-163 apple

    176-181 = ghosts on title screen

    206 = gate right
    207 = gate left
    208 = top right corner
    209 = top left corner
    210 = wall right
    211 = wall left
    212 = bottom right corner
    
    218 = top wall (right side of screen)
    219 = top wall (left side of screen)
    220 = bottom wall (right side of screen)
    221 = bottom wall (left side of screen)
    
    222 = top of object (right side of screen)
    223 = top of object (left side of screen)
    
    228 = bottom of object (right side of screen)
    229 = bottom of object (left side of screen)
    230 = top right corner of object
    231 = top left corner of object
    232 = right side of object
    233 = left side of object
    234 = bottom right of object
    235 = botom left of object

*/

static char textlayer[256]={
    1,1,1,1, 1,1,1,1, 1,1,0,0, 0,0,0,0,            /* 0-15 */
    0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,1,1,            /* 16-31 */
    1,1,1,1, 1,1,1,1, 1,1,1,1, 1,1,1,1,            /* 32-47 */
    1,1,1,1, 1,1,1,1, 1,1,1,1, 1,1,1,1,            /* 48-63 */
    1,1,1,1, 1,1,1,1, 1,1,1,1, 1,1,1,1,            /* 64-79 */
    1,1,1,1, 1,1,1,1, 1,1,1,1, 1,1,1,1,            /* 80-95 */
    0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0,            /* 96-111 */
    0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0,            /* 112-127 */
    0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0,            /* 128-143 */
    1,1,1,1, 1,1,1,1, 1,1,1,1, 1,1,1,1,            /* 144-159 */
    1,1,1,1, 0,0,0,0, 0,0,0,0, 0,0,0,0,            /* 160-175 */
    1,1,1,1, 1,1,0,0, 0,0,0,0, 0,0,0,0,            /* 176-191 */
    0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0,            /* 192-207 */
    0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0,            /* 208-223 */
    0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0,            /* 224-239 */
    0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0             /* 240-255 */
};

void drawtile3d(int tile, int color, int flipy, int flipx, int y, int x) {
    if (display_mode == CLASSIC) return;
    /* transparent tiles */
    if(tile==64 || tile>=252) return;

    x=SCREENWIDTH-x-8;

    glPushMatrix();

    /* text layer */
    if (textlayer[tile]
       || (y==192 && x==80 && tile==16) /* pill on title screen */
       || (y==208 && x==80 && tile==20) /* power pill on title screen */) {
      
      glLoadIdentity();
      glScalef(2.0/SCREENWIDTH, 2.0/SCREENWIDTH, 2.0/SCREENWIDTH);
    }
    
    color*=4;

    x=x+4-(28*4);
    y=-(y+4-(36*4));
    
    glTranslatef(x,y,0);
    if (flipx) FLIPX;
    if (flipy) FLIPY;
    draw_3d_tile(tile,color);
    glPopMatrix();
}


#if 0
static struct GfxLayout spritelayout =
{
    16,16,    /* 16*16 sprites */
    64,    /* 64 sprites */
    2,    /* 2 bits per pixel */
    { 0, 4 },    /* the two bitplanes for 4 pixels are packed into one byte */
    { 8*8, 8*8+1, 8*8+2, 8*8+3, 16*8+0, 16*8+1, 16*8+2, 16*8+3,
            24*8+0, 24*8+1, 24*8+2, 24*8+3, 0, 1, 2, 3 },
    { 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8,
            32*8, 33*8, 34*8, 35*8, 36*8, 37*8, 38*8, 39*8 },
    64*8    /* every sprite takes 64 bytes */
};
#endif



void video_sprite_init (void) {
    int sprite, x, y, ty;

    for (sprite=0;sprite<64;sprite++) {
        for (y=0; y<16; y++) {
            ty=y;
            if(y>7) ty+=24;
            for (x=0; x<4; x++) {
                sprites[sprite][x][y]=((spriterom[(sprite*64)+ty]>>x)&0x01)|
                                       ((spriterom[(sprite*64)+ty]>>(x+3))&0x02);

                sprites[sprite][x+4][y]=((spriterom[(sprite*64)+ty+24]>>x)&0x01)|
                                         ((spriterom[(sprite*64)+ty+24]>>(x+3))&0x02);

                sprites[sprite][x+8][y]=((spriterom[(sprite*64)+ty+16]>>x)&0x01)|
                                         ((spriterom[(sprite*64)+ty+16]>>(x+3))&0x02);

                sprites[sprite][x+12][y]=((spriterom[(sprite*64)+ty+8]>>x)&0x01)|
                                          ((spriterom[(sprite*64)+ty+8]>>(x+3))&0x02);
            }
        }
    }
}

#define MINX 0*8
#define MAXX 28*8-1
#define MINY 2*8
#define MAXY 34*8-1


void drawsprite(int sprite, int color, int flipy, int flipx, int y, int x) {
    Uint32 *pixels;

    int pixel;
    int i=0,j=0,ti=0,tj=0;
    int minx=0, maxx=16;
    int miny=0, maxy=16;

    if (magictile>=0 && magictile < 64) sprite=magictile;

    x=SCREENWIDTH-x-16;
    color*=4;

    /* clip */
    if(x < MINX) minx=MINX-x;
    if(x+16 > MAXX) maxx=MAXX-x;
    if(y+16 > MAXY) maxy=MAXY-y;
    if(y < MINY) miny=MINY-y;

    if(SDL_MUSTLOCK(video_buffer))
    SDL_LockSurface(video_buffer);

    pixels=(Uint32*)video_buffer->pixels;
    pixels+=((y+miny)*SCREENWIDTH) + x;

    for (i=miny;i<maxy;i++) {
        ti=flipy?i:(15-i);
        for (j=minx;j<maxx;j++) {
            tj=flipx?j:(15-j);
            pixel=colortable[color+sprites[sprite][ti][tj]];
            if(pixel)
                pixels[j]=paldata[pixel];
        }
        pixels+=SCREENWIDTH;
    }


    if(SDL_MUSTLOCK(video_buffer))
    SDL_UnlockSurface(video_buffer);

}


void drawsprite3d(int sprite, int color, int flipy, int flipx, int y, int x) {
    if (display_mode == CLASSIC) return;

    if (magictile>=0 && magictile < 64) sprite=magictile;

    /* disable pacman for 1st person view */
    if (display_mode == FIRST_PERSON && color == 9) return;

    if(color==0) return;

    x=SCREENWIDTH-x-16;

	
	if ((x+16)<0) return;
	if (x>MAXX) return;
	if (y>MAXY) return;
	if ((y+16)<0) return;

    color*=4;
    x=x+8-(28*4);
    y=-(y+8-(36*4));

    glPushMatrix();
    glTranslatef(x,y,0);
    if (flipx) FLIPX;
    if (flipy) FLIPY;

    draw_3d_sprite(sprite, color);
    glPopMatrix();
}

/*----------------------------------------------------------------------------*/
void video_emulate_video(void)
{
    int offs;

    for (offs = videoram_size - 1; offs > 0; offs--)
    {
        int mx,my,sx,sy;

        mx = offs % 32;
        my = offs / 32;

        if (my < 2)
        {
            if (mx < 2 || mx >= 30) continue; /* not visible */
            sx = my + 34;
            sy = mx - 2;
        }
        else if (my >= 30)
        {
            if (mx < 2 || mx >= 30) continue; /* not visible */
            sx = my - 30;
            sy = mx - 2;
        }
        else
        {
            sx = mx + 2;
            sy = my - 2;
        }

        if (flipscreen)
        {
            sx = 35 - sx;
            sy = 27 - sy;
        }

        if (dirtybuffer[offs]) {
            dirtybuffer[offs]=0;
            drawtile(videoram[offs],
                 colorram[offs] & 0x1f,
                 flipscreen, flipscreen,
                 sx*8,sy*8);
        }


        drawtile3d(videoram[offs],
                 colorram[offs] & 0x1f,
                 flipscreen, flipscreen,
                 sx*8,sy*8);
    }

    /* copy tile layer onto video_buffer */
    SDL_BlitSurface(tile_layer,&screenrect,video_buffer,&screenrect);

    /* Draw the sprites. Note that it is important to draw them exactly in this */
    /* order, to have the correct priorities. */
    for (offs = spriteram_size - 2;offs > 2*2;offs -= 2)
    {
        int sx,sy;

        sx = 272 - spriteram_2[offs + 1];
        sy = spriteram_2[offs] - 31;

        drawsprite(spriteram[offs] >> 2,
                spriteram[offs + 1] & 0x1f,
                spriteram[offs] & 1,spriteram[offs] & 2,
                sx,sy);

        drawsprite3d(spriteram[offs] >> 2,
                spriteram[offs + 1] & 0x1f,
                spriteram[offs] & 1,spriteram[offs] & 2,
                sx,sy);
    }
    /* In the Pac Man based games (NOT Pengo) the first two sprites must be offset */
    /* one pixel to the left to get a more correct placement */
    for (offs = 2*2;offs >= 0;offs -= 2)
    {
        int sx,sy;

        sx = 272 - spriteram_2[offs + 1];
        sy = spriteram_2[offs] - 31;

        drawsprite(spriteram[offs] >> 2,
                spriteram[offs + 1] & 0x1f,
                spriteram[offs] & 1,spriteram[offs] & 2,
                sx,sy + 1);

        drawsprite3d(spriteram[offs] >> 2,
                spriteram[offs + 1] & 0x1f,
                spriteram[offs] & 1,spriteram[offs] & 2,
                sx,sy + 1);
    }
}


void makedisplaylist (void) {
  if (glIsList(displayList))
    glDeleteLists(displayList, 1);

  displayList = glGenLists(1);
  glNewList(displayList, GL_COMPILE);

  glColor3f(1.0,1.0,1.0);

  glEnable(GL_TEXTURE_2D);

  glBindTexture(GL_TEXTURE_2D, pactexture);

  glBegin(GL_QUADS);
  
  /* upper right */
  glTexCoord2f(MAXTEXX, 0.0f); glVertex2f(MAXTEXY, 0.0f);
  /* lower right */
  glTexCoord2f(MAXTEXX, MAXTEXY); glVertex2f(MAXTEXY, -MAXTEXX);
  /* lower left */
  glTexCoord2f(0.0f, MAXTEXY); glVertex2f(0.0f, -MAXTEXX);
  /* upper left */
  glTexCoord2f(0.0f, 0.0f); glVertex2f(0.0f, 0.0f);

  glEnd();

  glDisable(GL_TEXTURE_2D);

  glEndList();
}


void setup_texture(void) {

  glGenTextures(1,&pactexture);

  glBindTexture(GL_TEXTURE_2D,pactexture);

  /* No borders */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

  /* Bilinear filtering */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

  glTexImage2D(GL_TEXTURE_2D,
               0,
           GL_RGBA,
           TEXTUREWIDTH,
           TEXTUREHEIGHT,
               0, GL_RGBA, GL_UNSIGNED_BYTE, video_buffer->pixels);

  makedisplaylist();

}

void drawpacman2D (void) {
  /* update texture */
  glBindTexture(GL_TEXTURE_2D, pactexture);
  glFlush();

  if(SDL_MUSTLOCK(video_buffer)) SDL_LockSurface(video_buffer);
  glTexSubImage2D(GL_TEXTURE_2D,
                  0, 0, 0,
                  SCREENWIDTH, SCREENHEIGHT, GL_RGBA,
                  GL_UNSIGNED_BYTE, video_buffer->pixels);

  if(SDL_MUSTLOCK(video_buffer)) SDL_UnlockSurface(video_buffer);
  
  /* draw pacman display */
  glCallList(displayList);
}


float rotx=-50.0f;
float rotz=-25.0f;

void video_mouse_rot(Sint16 xrel, Sint16 yrel) {
    rotx=rotx+(float)yrel/4.0f;
    rotz=rotz+(float)xrel/4.0f;
    
	if(rotx<-75.0f) rotx=-75.0f;
	if(rotx>0.0f) rotx=0.0f;
	if(rotz<-90.0f) rotz=-90.0f;
	if(rotz>90.0f) rotz=90.0f;

}


void video_isometric_view (void) {
  /* setup isometric view */
  glLoadIdentity();
  glTranslatef(0.0,0.0,-2.75);
  glRotatef(rotx, 1, 0 ,0);
  glRotatef(rotz, 0, 0, 1);
  glScalef(3.75/SCREENWIDTH,3.75/SCREENWIDTH,3.75/SCREENWIDTH);

}




/* horz =
   2c
   2e
   
   
   both =
   30
   
   vert =
   2d
   2f
   
   dying=34-3f
*/
static GLfloat olddirection=-90.0;
static GLfloat playerx=0.0;
static GLfloat playery=0.0;
static GLfloat direction=-90.0;
static GLfloat playerheight=-8.0;
void video_pacman_view (void) {
    int offs;
    int sprite = 0;
    int flipx = 1;
    int flipy = 1;
    int color = 0;
    int y = SCREENHEIGHT / 2;
    int x = SCREENWIDTH / 2;

    /* Draw the sprites. Note that it is important to draw them exactly in this */
    /* order, to have the correct priorities. */
    for (offs = spriteram_size - 2;offs > 2*2;offs -= 2)
    {
        int sx,sy;

	color  = spriteram[offs + 1] & 0x1f;
	if (color == 9) {
	    sx = 272 - spriteram_2[offs + 1];
            sy = spriteram_2[offs] - 31;
	    sprite = spriteram[offs] >> 2;
	    flipy  = spriteram[offs] & 1;
	    flipx  = spriteram[offs] & 2;
	    y = sx;
	    x = sy;
	    break;
	}
    }
    /* In the Pac Man based games (NOT Pengo) the first two sprites must be offset */
    /* one pixel to the left to get a more correct placement */
    if (color != 9) {
      for (offs = 2*2;offs >= 0;offs -= 2)
      {
          int sx,sy;
	  color  = spriteram[offs + 1] & 0x1f;
	  if (color == 9) {
              sx = 272 - spriteram_2[offs + 1];
              sy = spriteram_2[offs] - 31;

              sprite = spriteram[offs] >> 2;

	      flipy  = spriteram[offs] & 1;
	      flipx  = spriteram[offs] & 2;
	      y = sx;
	      x = sy + 1;
	      break;
	  }
      }
    }

    if (color == 9) {
	x=SCREENWIDTH-x-16;
	playerx=2.0f*((((float)(x+8))/((float)SCREENWIDTH))-0.5)*SCREENRATIO;
	playery=-2.0f*((((float)(y+8))/((float)SCREENHEIGHT))-0.5)/SCREENRATIO;

	playerx=(28*4)-x-8;
	playery=(36*4)-y-8;

	/* face in pacmans directions (north) */
	if(sprite<0x30) {
	   if (sprite & 0x01) {
	     if (flipy) {
	       /* north */
	       direction=0;
	       input_direction(NORTH);
	     } else {
	       /* south */
	       direction=180;
	       input_direction(SOUTH);
	     }
	   } else {
	     if (flipx) {
	       /* west */
	       direction=270;
	       input_direction(WEST);
	     } else {
	       /* east */
	       direction=90;
	       input_direction(EAST);
	     }
	   }
	}

	/* change height on death */
	if (sprite<0x34 || sprite >0x3f) {
	  playerheight= -8.0;
	} else {
	  playerheight= -8.0 * (0x40-sprite) / 12.0f;
	}
    }

    /* scale for sprite and viewer */
    glScalef(3.0*8.0,3.0*8.0,3.0*8.0);
    glTranslatef(0,0,-8.0);

    /* chase mode */
    if (display_mode == CHASE_VIEW) {
        /* move slightly behind and up */
        glTranslatef(0,-8,-56.0);
        /* point down a little */
        glRotatef(20.0,1,0,0);
    }

    if (olddirection != direction) {
	GLfloat diff=direction-olddirection;

	if (diff>180.0 || diff<-180.0) diff=-diff;
	if (diff<0) diff=-9.0;
	else diff=9.0;
    	direction=olddirection+diff;
    	if (direction<0) direction+=360.0;
    	if (direction>360) direction-=360.0;
        olddirection=direction;
    }


    /* rotate to pacmans direction */
    glRotatef(direction,0,1,0);

    /* move to pacmans position */
    glTranslatef(playerx,playerheight,playery);

    /* correct plane */
    glRotatef(-90,1,0,0);


}


void video_draw_screen (void) {

  /* Clear the color and depth buffers. */
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  /* We don't want to modify the projection matrix. */
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity( );

  /* setup view */
  if (display_mode >= FIRST_PERSON) {
    video_pacman_view();
  } else if (display_mode == THIRD_PERSON) {
    video_isometric_view();
  }

  /* emulate pacman display hardware */
  video_emulate_video();


  /* draw 2D pacman display */
  if (display_mode != CLASSIC) {
    glLoadIdentity();
    glTranslatef(-1.40f,1.00f,0.5f);
    glScalef(0.75f,0.75f,0.75f);
  } else {
    glTranslatef(-0.9,1.4,0.0);
    glScalef(3.2,3.2,1.0);
  }
  drawpacman2D();

  /* render frame */
  SDL_GL_SwapBuffers();
}



void snapshot_init(void)
{
    /* This function sets up snapshot number */
    char filename[30];
    FILE *fp;

    snap_no=0;

    sprintf(filename,"snap%04d.bmp", snap_no);

    while((fp=fopen(filename,"r"))!=NULL)
    {
        fclose(fp);
        snap_no++;
        sprintf(filename,"snap%04d.bmp", snap_no);
    }
}




void video_save_snapshot(void)
{
    SDL_Surface *temp;
    unsigned char *pixels;
    int i;
    char filename[30];
    FILE *fp;

    sprintf(filename,"snap%04d.bmp", snap_no);
    while((fp=fopen(filename,"r"))!=NULL)
    {
        fclose(fp);
        snap_no++;
        sprintf(filename,"snap%04d.bmp", snap_no);
    }

    temp = SDL_CreateRGBSurface(SDL_SWSURFACE, 
                                width, height, 32,
				rmask,gmask,bmask,amask);
    if (temp == NULL) return;

    pixels = malloc(4 * width * height);
    if (pixels == NULL) {
            SDL_FreeSurface(temp);
            return;
    }

    glReadPixels(0, 0, width, height,
                 GL_RGBA, GL_UNSIGNED_BYTE, pixels);

    for (i=0; i<height; i++)
            memcpy(((char *) temp->pixels) + temp->pitch * i, 
                   pixels + 4*width * (height-i-1),
                   width*4);
    free(pixels);

    SDL_SaveBMP(temp, filename);
    SDL_FreeSurface(temp);

}

